%% Vessel Dynamics
function [eta,nu,t,stop] = simplifiedVesselModelSim1(eta0,nu0,tau0,v_c_n0,MRB,MA,Bv,Dp,G,rho_w,Ax,lpp,T,B,WP,WPk0, fm_start, fm_duration, rudderLoss, powerLoss, powerFrac, totalLoss)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Nonlinear Unified Seakeeping and Maneuvering Model  
% without Fluid Memory,based on Perez & Fossen 2017.
% 
% The function also includes a PID autopilot controller.
%
% Author: Markus Fossdal 2018.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Simulation parameters
 Ts_sim = 0.1;
 Tlength = 60;
 Tsim = Tlength/Ts_sim;
 
% Parameters
M = MRB + MA; %Rigid-body mass + added mass matrix
M_inv = inv(M);


L = [0 0 0 0 0 0; %Selection matrix
     0 0 0 0 0 1;
     0 0 0 0 -1 0;
     0 0 0 0 0 0;
     0 0 0 0 0 0; 
     0 0 0 0 0 0];

R_zc = Rzyx(eta0(4),eta0(5),eta0(6));
nu_c0 = [R_zc'*v_c_n0; 0 ; 0 ; 0]; %Body current

nCs = 20; %Number of vessel cross sections

%Controller
i_term_N = 0; %integral term initial value

% Memory allocation
nu = zeros(6,Tsim);
nu_c = zeros(6,Tsim);
nu_r = zeros(6,Tsim);
eta = zeros(6,Tsim);
t = zeros(1,Tsim);
tau_N_temp = zeros(1,Tsim);

% Initial conditions
eta(:,1) = eta0;
nu(:,1) = nu0;
nu_c(:,1) = nu_c0;
t(1) = 0;
WPk = WPk0;
% Computations
%Solves combined forward and backward Euler integration (Fossen 2011, appendix B)
for k = 1:Tsim-1
    
    %Current rotation from NED to Body
    R_zc = Rzyx(eta(4,k),eta(5,k),eta(6,k));
    nu_c(:,k) = [R_zc'*v_c_n0; 0 ; 0 ; 0]; %Body current
    
    %Relative velocity
    nu_r(:,k) = nu(:,k) - nu_c(:,k); %relative nu
    U_r = nu_r(1,k); %relative surge velocity
    
    %Coriolis matrices rigid-body and added mass.
    CRB = U_r * MRB * L;
    CA = U_r * MA * L;
    
    %Quadratic surge resistance
    Cx = 1;
    qsr_x = -0.5*rho_w*Cx*abs(nu_r(1,k))*nu_r(1,k)*Ax;
    tau_qsr = [qsr_x 0 0 0 0 0]';
    
    %Cross-flow drag
    cs = -(lpp/2):(lpp/(nCs-1)):(lpp/2);
    dx = lpp/(nCs-1);
    Y_cf = 0;
    N_cf = 0;
    for i=1:nCs
        Y_cf = Y_cf + (-0.5*rho_w * T*Hoerner(B,T)*abs( nu_r(2,k)+cs(i)*nu(6) ) * ( nu_r(2,k)+cs(i)*nu(6,k) ) )*dx;
        N_cf = N_cf + (-0.5*rho_w * T*Hoerner(B,T)*cs(i)*abs( nu_r(2,k)+cs(i)*nu(6,k) ) * ( nu_r(2,k)+cs(i)*nu(6,k) ) )*dx;
    end
    tau_cf = [0 Y_cf 0 0 0 N_cf]'; %Cross-flow drag, Y and N

    %Waypoint Guidance
    [chi_d,WPk1] = guidance([eta(1,k);eta(2,k)], WP, lpp,WPk);
    WPk = WPk1;
   
    %Sideslip compensation
    [psi_d] = sideSlipComp(chi_d,nu(:,k));
     
   
    %Input calculation (tau)
    psi = eta(6,k); %heading
    r = nu(6,k); %yaw-rate
    
    Kp_N = 0.8*MRB(6,6); %Proportional gain 
    Ki_N = 0; %Integral gain
    Kd_N = 0.6*MRB(6,6)*10; %Derivative gain
    
    e_N = rad2pipi(psi_d-psi); %PID Error
    
    p_term_N = Kp_N * e_N; %PID proportional contribution
    i_term_N = i_term_N + Ki_N * e_N; %PID integral contribution
    d_term_N = -Kd_N * r; %PID derivative contribution
    
    tau_N = p_term_N + i_term_N + d_term_N; %PID Control
    
    %Failure mode logic
    if rudderLoss && powerLoss
        stop = true;
        error('Rudder- and Power loss cannot be activated at the same time.');
    else
        stop = false;
    end
    
    if rudderLoss
        %Rudder loss logic
        if t(k)== fm_start
            tau_N_temp(k:k+((20+45)/Ts_sim)) = tau_N;
        end
        
        if t(k) > fm_start && t(k) <= (fm_start+fm_duration)
            tau_N_used = tau_N_temp(k);
        else
            tau_N_used = tau_N;
        end
        
        tau = [1*tau0(1,1) 0 0 0 0 tau_N_used]';
    
    elseif powerLoss
        %Power loss logic
        if t(k) > fm_start && t(k) <= (fm_start+fm_duration)
            tau = [powerFrac*tau0(1,1) 0 0 0 0 tau_N]';
        else
            tau = [tau0(1,1) 0 0 0 0 tau_N]';
        end
    elseif totalLoss
        %Lose both rudder and power
        if t(k) > fm_start && t(k) <= (fm_start+fm_duration)
            tau = [0 0 0 0 0 0]';
        else
            tau = [tau0(1,1) 0 0 0 0 tau_N]';
        end
        
    else
        tau = [tau0(1,1) 0 0 0 0 tau_N]';
    end
    
    
    %Equations of motion
    nu(:,k+1) = nu(:,k) + Ts_sim*M_inv*(tau - CRB*nu(:,k) -CA*nu_r(:,k) - (Bv+Dp)*nu_r(:,k) - G*eta(:,k) + tau_qsr + tau_cf);
    [J,~,~] = eulerang(eta(4,k),eta(5,k),eta(6,k));
    eta(:,k+1) = eta(:,k) + Ts_sim*J*nu(:,k+1);
    t(k+1) = t(k) + Ts_sim;
        
end %simulation

end

 
 %% Guidance Law
function [chi_d,k1] = guidance(pos, wayPoints, L_ship,k)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Lookahead-based steering law, taken from Fossen (2011).
%
% Author: Markus Fossdal 2018.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Design lengths for guidance system
R_accept = L_ship*2;
R_los = L_ship*3;

% Loading the waypoints
xPos = wayPoints(1,:);
yPos = wayPoints(2,:);

% Correcting the first waypoint if zero
if xPos(1) == 0
    xPos(1) = -0.00001;
end
if yPos(1) == 0
    yPos(1) = -0.00001;
end

% If still not reached the first waypoint
if k == 0
    xk = 0;
    yk = 0;
else
    xk = xPos(k);
    yk = yPos(k);
end

% The next waypoint
xk1 = xPos(k+1);
yk1 = yPos(k+1);

% Finding the path-tangential angle
chiP = atan2(yk1-yk,xk1-xk);

% Finding the cross-track error
e = -(pos(1)-xk)*sin(chiP) + (pos(2)-yk)*cos(chiP);

% Correcting e if larger than R_los
if abs(e) > R_los
    e = R_los-1;
end

% Using a constant lookahead distance if not reached the first waypoint
if k == 0
    delta = L_ship;
else
    delta = sqrt(R_los^2 - e^2);
end

% Caluclating the desired course
chi_d = rad2pipi(chiP + atan(-e/delta));

% Switch to next waypoint if within circle of acceptance
if (xk1-pos(1))^2 + (yk1-pos(2))^2 < R_accept^2
    if k < length(xPos)-1
        k1 = k+1;
    else
        k1 = k;
    end    
else
    k1 = k;
end
end

%% Sideslip Compensation
function [psi_d] = sideSlipComp(chi_d,nu)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Sideslip compensation algorithm, from Fossen (2011).
%
% Author: Markus Fossdal 2018.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
U = sqrt(nu(1)*nu(1)+nu(2)*nu(2)); %Vessel speed

if U == 0 
    U = 0.001;
end

beta = asin(nu(2)/U); %Sideslip angle
psi_d = chi_d-beta; %Corrected heading input
end


%% Consequence Level Evaluation
function [level,vioLvl] = levelCheck(eta,t,obstacles,obstLevels, noObst)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Consequence level evaluation, from Fossdal (2018).
%
% Author: Markus Fossdal 2018.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%Initialize level at zero
level = 0;
maxLevel = 0;

%Declare empty distance vector
dist2obj = zeros(1,size(eta,2));

%Declare empty minimum distance vector
minDist = zeros(1,size(obstacles,2));


%1) For loop through obstacles and save the smallest distance to the different
%objects
for i = 1:size(obstacles,2)
    for j = 1:size(eta,2)
        %Calculate scalar distance from vessel to obstacles
        dist2obj(j) = sqrt((obstacles(2,i)-eta(2,j))^2 + (obstacles(1,i)-eta(1,j))^2);   
    end 
   %Vector containing the closest distance during a simulation to the obstacles
   minDist(i) = min(dist2obj);
end

%2) Check if these distances violate the largest level zone.
%Declare relevant index values
relIdx = zeros(1,size(obstacles,2));

%Index to closest obstacle
if size(obstacles,2) > 1
    for i=1:size(obstacles,2)
        if minDist(i) < obstLevels(1,1)
            relIdx(i) = 1;
        else
            relIdx(i) = 0;
        end
    end
else
    relIdx = 1;
end

% 3) Investigate which levels that are violated. Output highest level to
% online system.

%Declare violation level vector
vioLvl = zeros(size(t,2),size(obstacles,2));

for i=1:size(obstacles,2)
    if relIdx(i)
        for j = 1:size(eta,2)
            dist2obj(j) = sqrt((obstacles(2,i)-eta(2,j))^2 + (obstacles(1,i)-eta(1,j))^2);
            for k = 1:noObst
               if dist2obj(j) < obstLevels(k)
                  vioLvl(j,i)=k;
               end
            end
        end
    end   
end

maxLevel = max(max(vioLvl));


%Output max level violated in simulation.
if maxLevel > 0
    level = maxLevel;
elseif maxLevel == 0
    level = -1;
else
    level = 0;
end


end


%% Risk evaluation function
function [riskIn,riskInMax,fmIdx] = riskAnalysis(prob,CL1,CL2,CL3,CL4)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Total risk level evaluation, from Fossdal (2018).
%
% Author: Markus Fossdal 2018.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Put CL in vector
CL = [CL1,CL2,CL3,CL4];

%CL mapped
CL_m = zeros(1,size(CL,2));

%Map consequence levels
for i=1:size(CL,2)
    if CL(1,i) == -1
        CL_m(1,i) = 0;
    elseif CL(1,i) == 1
        CL_m(1,i) = 0.33333;
    elseif CL(1,i) == 2
        CL_m(1,i) = 0.66667;
    elseif CL(1,i) == 3
        CL_m(1,i) = 1;
    end
end

%Calculate risk levels
r = zeros(1,size(CL,2));
for i=1:size(r,2)
    r(i) = prob(i)*CL_m(i);
end

%Calculate max risk
r_max = 1*sum(prob);

%Aquire risk indicator and fm with the largest risk.
fmIdx = 0;
[riskInMax,fmIdx] = max(r);
riskIn = sum(r)/r_max;

if max(r) == 0
    fmIdx = 0;
end

end